#include "cardappz.h"
#include "hardware/springprox/springprox.h"
#include <assert.h>
#include <stdio.h>
#include <conio.h>

static int DesfireSamTransmitEx(CARDAPPZ_CTX_ST * ctx, unsigned char cmd[],
                                int cmdlen, unsigned char rsp[], int *rsplen)
{
  SWORD rc;

  assert(ctx != NULL);

  if (ctx->trace >= DUMPMORE)
  {
    int i;

    printf("Card<");
    for (i = 0; i < cmdlen; i++)
      printf("%02X", cmd[i]);
    printf("\n");
  }

  /* SCard transmit using SpringProx reader  */
  /* not difficult to change for PC/SC mode */
  if (rsplen != NULL)
  {
    WORD t = (WORD) * rsplen;

    rc =
      SPROX_Card_Exchange(ctx->desfire_sam.slot, cmd, (WORD) cmdlen, rsp, &t);
    *rsplen = t;
  } else
  {
    rc =
      SPROX_Card_Exchange(ctx->desfire_sam.slot, cmd, (WORD) cmdlen, rsp,
                          NULL);
  }

  if (ctx->trace >= DUMPMORE)
  {
    int i;

    printf("Card>");
    if ((rsp != NULL) && (rsplen != NULL))
      for (i = 0; i < *rsplen; i++)
        printf("%02X", rsp[i]);
    printf("\n");
  }

  return rc;
}

static int DesfireSamTransmit(CARDAPPZ_CTX_ST * ctx, unsigned char CLA,
                              unsigned char INS, unsigned char P1,
                              unsigned char P2, const unsigned char data[],
                              int LC, int LE, unsigned char rsp[],
                              int *rsplen, unsigned short *status)
{
  unsigned char io_buffer[128];
  int o_len, i_len;
  unsigned char sw[2];
  int rc;

  if ((LC + 6) >= sizeof(io_buffer))
    return CARDAPPZ_BUFFER_OVERFLOW;

  /* Build the ISO 7816-4 APDU */
  io_buffer[0] = CLA;
  io_buffer[1] = INS;
  io_buffer[2] = P1;
  io_buffer[3] = P2;

  o_len = 4;
  if (data != NULL)
  {
    io_buffer[4] = LC;
    memcpy(&io_buffer[5], data, LC);
    o_len = 5 + LC;
  }
  if (LE >= 0)
  {
    io_buffer[o_len] = (unsigned char) LE;
    o_len++;
  }

  i_len = sizeof(io_buffer);
  if (rsplen != NULL)
    if (i_len > (*rsplen + 1))
      i_len = (*rsplen + 1);

  rc = DesfireSamTransmitEx(ctx, io_buffer, o_len, io_buffer, &i_len);
  if (rc)
    return rc;

  if (i_len < 2)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;

  if (rsp != NULL)
  {
    memcpy(rsp, io_buffer, i_len);
    if (rsplen != NULL)
      *rsplen = i_len;
  }

  sw[0] = io_buffer[i_len - 2];
  sw[1] = io_buffer[i_len - 1];

  if (ctx->trace >= TRACE)
    if ((sw[0] != 0x90) || (sw[1] != 0x00))
      printf("DesfireSam: card status %02X%02X\n", sw[0], sw[1]);

  if (status != NULL)
  {
    *status = sw[0] << 8;
    *status |= sw[1];
  }

  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMAuthenticateHost
 *
 * NAME
 *   DesfireSAMAuthenticateHost
 *
 * DESCRIPTION
 *   Desfire SAM mutual authentication command (CLA=80 INS=A4 P1=00)
 *   (functional specification 1.0, chapter 4.2.7)
 *   This generate a session key.
 *   If SamUid is provided, the authentication uses Desfire SAM key
 *   diversification algorithm
 *
 * SYNOPSIS
 *   int DesfireSAMAuthenticateHost(CARDAPPZ_CTX_ST *ctx,
 *                                  unsigned char KeyIndex,
 *                                  const unsigned char KeyValue[16],
 *                                  const unsigned char SamUid[7]);
 *
 **/
int DesfireSAMAuthenticateHost(CARDAPPZ_CTX_ST * ctx,
                               unsigned char KeyIndex,
                               const unsigned char KeyValue[16],
                               const unsigned char SamUid[7])
{
  unsigned char cmd[17];
  unsigned char rsp[10];
  unsigned char sess_key[16];
  unsigned short status;
  int is_tdes;
  int rsplen;
  int rc;
  int i;
  unsigned char KeyVersion = 0;
  unsigned char RndA[8], RndB[8];
  unsigned char P1;
  TDES_CTX_ST tdes_ctx;

  assert(ctx != NULL);
  if (ctx->trace >= TRACE)
  {
    if (SamUid == NULL)
      printf("DesfireAuthenticate with key %d\n", KeyIndex);
    else
      printf("DesfireAuthenticate with key %d (diversified)\n", KeyIndex);
  }

  if (KeyValue == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  /* Loose current authentication status */
  ctx->card_context.desfire.auth_status = 0;

  /* 3-DES or single DES ? */
  is_tdes = memcmp(&KeyValue[0], &KeyValue[8], 8);

  /* Preload 3-DES context with key */
  TDES_Init(&tdes_ctx, &KeyValue[0], &KeyValue[8], &KeyValue[0]);

  /* Is SamUid is provided, we are to diversicate the key */
  if (SamUid != NULL)
  {
    /* First block <- SamUid XOR first part of master key */
    sess_key[0] = 0x88;
    memcpy(&sess_key[1], SamUid, 7);
    for (i = 0; i < 8; i++)
      sess_key[i] ^= KeyValue[i];

    /* Cipher first block, this is first part of diversified key */
    TDES_Encrypt(&tdes_ctx, &sess_key[0]);

    /* Second block <- ciphered first block XOR second part of master key */
    for (i = 0; i < 8; i++)
      sess_key[8 + i] = sess_key[i] ^ KeyValue[8 + i];

    /* Cipher second block, this is second part of diversified key */
    TDES_Encrypt(&tdes_ctx, &sess_key[8]);

    /* Reload 3-DES context with diversified key */
    TDES_Init(&tdes_ctx, &sess_key[0], &sess_key[8], &sess_key[0]);

    P1 = 0x01;
  } else
  {
    P1 = 0x00;
  }

  /* Get 8 random bytes as RndA */
  for (i = 0; i < 8; i++)
    RndA[i] = rand();
  /* Note : C library rand() is a WEAK pseudo-random number */
  /* generator, use a strong PRNG instead when available !  */

  /* Build command, we want 8+2 bytes in reply */
  cmd[0] = KeyIndex;
  cmd[1] = KeyVersion;
  rsplen = 10;
  rc =
    DesfireSamTransmit(ctx, 0x80, 0xA4, P1, 0x00, cmd, 2, 0, rsp, &rsplen,
                       &status);
  if (rc)
    return rc;
  if (status != 0x90AF)
    return -status;             /* Wrong status code */
  if (rsplen != 10)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;

  /* We've received Ek(RndB), decipher it */
  memcpy(RndB, &rsp[0], 8);
  TDES_Decrypt(&tdes_ctx, RndB);

  /* Create Ek(RndA || RndB') */
  memcpy(&cmd[0], RndA, 8);
  memcpy(&cmd[8], &RndB[1], 7);
  cmd[15] = RndB[0];
  TDES_Encrypt_CBC_S(&tdes_ctx, &cmd[0], 2);

  /* We want 8+2 bytes in reply */
  rsplen = 10;
  rc =
    DesfireSamTransmit(ctx, 0x80, 0xA4, 0x00, 0x00, cmd, 16, 0, rsp, &rsplen,
                       &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;             /* Wrong status code */
  if (rsplen != 10)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;

  /* Retrieve RndA' in answer */
  TDES_Decrypt(&tdes_ctx, &rsp[0]);
  if (memcmp(&rsp[0], &RndA[1], 7) || (rsp[7] != RndA[0]))
    return CARDAPPZ_CARD_AUTH_FAILURE;  /* SAM has not the appropriate key !!! */

  if (is_tdes)
  {
    /* Build the 3-DES session key */
    memcpy(&sess_key[0], &RndA[0], 4);
    memcpy(&sess_key[4], &RndB[0], 4);
    memcpy(&sess_key[8], &RndA[4], 4);
    memcpy(&sess_key[12], &RndB[4], 4);
  } else
  {
    /* Build the single DES session key */
    memcpy(&sess_key[0], &RndA[0], 4);
    memcpy(&sess_key[4], &RndB[0], 4);
    memcpy(&sess_key[8], &RndA[0], 4);
    memcpy(&sess_key[12], &RndB[0], 4);
  }

  TDES_Init(&ctx->desfire_sam.tdes_ctx, &sess_key[0], &sess_key[8],
            &sess_key[0]);
  ctx->desfire_sam.auth_status = 1;
  ctx->desfire_sam.auth_key = KeyIndex;

  /* OK */
  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMChangeKeyEntry
 *
 * NAME
 *   DesfireSAMChangeKeyEntry
 *
 * DESCRIPTION
 *   Desfire SAM mutual authentication command (CLA=80 INS=C1)
 *   (functional specification 1.0, chapter 4.2.1)
 *   APDU data are encrypted (3-DES CBC) with the session key 
 *   previously generated by a successfull DesfireSAMAuthenticateHost
 *
 * SYNOPSIS
 *   int DesfireSAMChangeKeyEntry(CARDAPPZ_CTX_ST *ctx,
 *                                unsigned char   KeyIndex,
 *                                unsigned char   ProMas,
 *                        [input] unsigned char   KeyValueA[16],
 *                        [input] unsigned char   KeyValueB[16],
 *                        [input] unsigned char   KeyValueC[16],
 *                                unsigned long   PiccAid,
 *                                unsigned char   PiccKeyIndex,
 *                                unsigned char   KeyNoCEK,
 *                                unsigned char   KeyVCEK,
 *                                unsigned char   RefNoKUC,
 *                                unsigned short  Settings,
 *                                unsigned char   SamUID[7]);
 *
 **/
int DesfireSAMChangeKeyEntry(CARDAPPZ_CTX_ST * ctx,
                             unsigned char KeyIndex,
                             unsigned char ProMas,
                             unsigned char KeyValueA[16],
                             unsigned char KeyValueB[16],
                             unsigned char KeyValueC[16],
                             unsigned long PiccAid,
                             unsigned char PiccKeyIndex,
                             unsigned char KeyNoCEK,
                             unsigned char KeyVCEK,
                             unsigned char RefNoKUC,
                             unsigned short Settings, unsigned char SamUid[7])
{
  int rc;
  unsigned char data[72] = { 0 };
  unsigned char len;
  unsigned short status;

  assert(ctx != NULL);

  /* Set data */
  if (KeyValueA != NULL)
  {
    memcpy(&data[0], KeyValueA, 16);
  }
  if (KeyValueB != NULL)
  {
    memcpy(&data[16], KeyValueB, 16);
  }
  if (KeyValueC != NULL)
  {
    memcpy(&data[32], KeyValueA, 16);
  }
  data[48] = (unsigned char) (PiccAid % 0x00000100);
  PiccAid /= 0x00000100;
  data[49] = (unsigned char) (PiccAid % 0x00000100);
  PiccAid /= 0x00000100;
  data[50] = (unsigned char) (PiccAid % 0x00000100);
  data[51] = PiccKeyIndex;
  data[52] = KeyNoCEK;
  data[53] = KeyVCEK;
  data[54] = RefNoKUC;
  data[55] = (unsigned char) (Settings % 0x0100);
  data[56] = (unsigned char) (Settings / 0x0100);
  len = 57;
  if (SamUid != NULL)
  {
    memcpy(&data[57], SamUid, 7);
    len += 7;
  }

  /* Add CRC */
  SPROX_ComputeIso14443ACrc(&data[len], &data[0], len);
  len += 2;
  len = 8 * ((7 + len) / 8);

  if (ctx->trace >= DUMPMORE)
  {
    int i;

    printf("len=%02X (%d blocks)\n", len, len / 8);
    for (i = 0; i < len; i++)
      printf("%02X", data[i]);
    printf("\n");
  }

  /* Encrypt the data */
  TDES_Encrypt_CBC_S(&ctx->desfire_sam.tdes_ctx, data, len / 8);

  /* Send the APDU */
  rc =
    DesfireSamTransmit(ctx, 0x80, 0xC1, KeyIndex, ProMas, data, len, -1,
                       NULL, NULL, &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;             /* Wrong status word */

  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMGetKeyEntry
 *
 * NAME
 *   DesfireSAMGetKeyEntry
 *
 * DESCRIPTION
 *   Retrieve a key entry in Desfire SAM (CLA=80 INS=64)
 *   (functional specification 1.0, chapter 4.2.2)
 *
 * SYNOPSIS
 *   int DesfireSAMGetKeyEntry(CARDAPPZ_CTX_ST *ctx,
 *                             unsigned char   KeyIndex,
 *                    [output] unsigned char   KeyEntry[12]);
 *
 **/
int DesfireSAMGetKeyEntry(CARDAPPZ_CTX_ST * ctx,
                          unsigned char KeyIndex, unsigned char KeyEntry[12])
{
  int rc;
  unsigned char rsp[64];
  int rsplen = sizeof(rsp);
  unsigned short status;

  rc =
    DesfireSamTransmit(ctx, 0x80, 0x64, KeyIndex, 0x00, NULL, 0, 0, rsp,
                       &rsplen, &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;             /* Wrong status word */
  if (rsplen != 14)
    return CARDAPPZ_CARD_CONTENT_ERROR;

  if (KeyEntry != NULL)
    memcpy(KeyEntry, rsp, 12);

  return CARDAPPZ_OK;
}

int DesfireSAMGetVersion(CARDAPPZ_CTX_ST * ctx, unsigned char Version[28],
                         unsigned char SamUid[7])
{
  int rc, i;
  unsigned char rsp[64];
  int rsplen = sizeof(rsp);
  unsigned short status;

  rc =
    DesfireSamTransmit(ctx, 0x80, 0x60, 0x00, 0x00, NULL, 0, 0, rsp, &rsplen,
                       &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;             /* Wrong status word */

  printf("VERSION=");
  for (i = 0; i < rsplen - 2; i++)
    printf("%02X", rsp[i]);
  printf("\n");

  if (rsplen < 30)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;

  if (Version != NULL)
    memcpy(Version, rsp, 28);
  if (SamUid != NULL)
    memcpy(SamUid, &rsp[14], 7);


  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMSelectKey
 *
 * NAME
 *   DesfireSAMSelectKey
 *
 * DESCRIPTION
 *   Select a key in Desfire SAM for further MAC or ciphering actions
 *   (CLA=80 INS=A4 P1=40) (functional specification 1.0, chapter 4.2.7)
 *
 * SYNOPSIS
 *   int DesfireSAMSelectKey(CARDAPPZ_CTX_ST *ctx,
 *                           unsigned char   KeyIndex,
 *                           unsigned char KeyVersion);
 *
 **/
int DesfireSAMSelectKey(CARDAPPZ_CTX_ST * ctx, unsigned char KeyIndex,
                        unsigned char KeyVersion)
{
  unsigned char cmd[2];
  unsigned short status;
  int rc;

  assert(ctx != NULL);

  if (ctx->trace >= TRACE)
    printf("DesfireSAMSelectKey %d %02X\n", KeyIndex, KeyVersion);

  /* Build command */
  cmd[0] = KeyIndex;
  cmd[1] = KeyVersion;

  /* Send APDU */
  rc =
    DesfireSamTransmit(ctx, 0x80, 0xA4, 0x04, 0x00, cmd, 2, 0, NULL, NULL,
                       &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;             /* Wrong status code */

  /* OK */
  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMAuthenticatePICC_1
 *
 * NAME
 *   DesfireSAMAuthenticatePICC_1
 *
 * DESCRIPTION
 *   Implementation of Desfire PICC authentication using Desfire SAM 
 *   command (CLA=80 INS=0A), step 1
 *   (functional specification 1.0, chapter 4.3.1)
 *   The selected key is diversified by the SAM using Desfire PICC serial
 *   number (card's UID)
 *   PiccChallenge is the challenge sent by the Desfire card in reply
 *   to the DesfireAuthenticate_1 command
 *   SamCryptogram (SAM's answer) must be forwarded as is to the PICC.
 *
 * SYNOPSIS
 *   int DesfireSAMAuthenticatePICC_1(CARDAPPZ_CTX_ST *ctx,
 *                                    unsigned char   KeyIndex,
 *                                    unsigned char   KeyVersion,
 *                           [input]  unsigned char   PiccChallenge[8]);
 *                           [output] unsigned char   SamCryptogram[16]);
 *
 **/
int DesfireSAMAuthenticatePICC_1(CARDAPPZ_CTX_ST * ctx,
                                 unsigned char KeyIndex,
                                 unsigned char KeyVersion,
                                 unsigned char PiccChallenge[8],
                                 unsigned char SamCryptogram[16])
{
  unsigned char cmd[18];
  unsigned char rsp[18];
  unsigned short status;
  int rsplen;
  int rc;

  assert(ctx != NULL);

  if (ctx->trace >= TRACE)
    printf("DesfireSAMAuthenticatePICC_1 with key %d\n", KeyIndex);

  if (PiccChallenge == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (SamCryptogram == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  /* Build command */
  cmd[0] = KeyIndex;
  cmd[1] = KeyVersion;
  memcpy(&cmd[2], PiccChallenge, 8);
  memset(&cmd[10], 0x88, 8);    /* Padding with 0x88 on the left of serial number */
  if (ctx->card_type & CARD_TYPE_14443_A)
  {
    if (ctx->card_ident.a.snrlen > 8)
      memcpy(&cmd[10], ctx->card_ident.a.snr, 8);
    else
      memcpy(&cmd[18 - ctx->card_ident.a.snrlen], ctx->card_ident.a.snr,
             ctx->card_ident.a.snrlen);
  } else if (ctx->card_type & CARD_TYPE_14443_B)
  {
    memcpy(&cmd[14], ctx->card_ident.b.atq, 4);
  } else
    return CARDAPPZ_EXCEPTION;

  /* Send APDU, we expect 16+2 bytes in reply */
  rsplen = 18;
  rc =
    DesfireSamTransmit(ctx, 0x80, 0x0A, 0x01, 0x00, cmd, 18, 0, rsp, &rsplen,
                       &status);
  if (rc)
    return rc;
  if (status != 0x90AF)
    return -status;             /* Wrong status code */
  if (rsplen != 18)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;

  /* We've received SAM cryptogram */
  memcpy(SamCryptogram, &rsp[0], 16);

  /* OK */
  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMAuthenticatePICC_2
 *
 * NAME
 *   DesfireSAMAuthenticatePICC_2
 *
 * DESCRIPTION
 *   Implementation of Desfire PICC authentication using Desfire SAM 
 *   command (CLA=80 INS=0A), step 2
 *   (functional specification 1.0, chapter 4.3.1)
 *   PiccCryptogram is the Desfire card answer to the DesfireAuthenticate_2
 *   command
 *   If result is OK,
 *   - the Desfire card can be trusted (it actually knows the key)
 *   - the Desfire card trusts the host application
 *   - a shared session key has been established between the SAM and the
 *     card (but unknown to the host application) to protect (MAC or cipher)
 *     further communications
 *
 * SYNOPSIS
 *   int DesfireSAMAuthenticatePICC_2(CARDAPPZ_CTX_ST *ctx,
 *                           [input]  unsigned char   PiccCryptogram[8]);
 *
 **/
int DesfireSAMAuthenticatePICC_2(CARDAPPZ_CTX_ST * ctx,
                                 unsigned char PiccCryptogram[8])
{
  unsigned short status;
  int rc;

  assert(ctx != NULL);

  if (ctx->trace >= TRACE)
    printf("DesfireSAMAuthenticatePICC_2\n");

  if (PiccCryptogram == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  /* Send APDU, we expect only a status word in reply */
  rc =
    DesfireSamTransmit(ctx, 0x80, 0x0A, 0x00, 0x00, PiccCryptogram, 8, -1,
                       NULL, NULL, &status);
  if (rc)
    return rc;
  if (status != 0x9000)
    return -status;             /* Wrong status code */

  /* OK */
  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMVerifyMAC
 *
 * NAME
 *   DesfireSAMVerifyMAC
 *
 * DESCRIPTION
 *   Use the Desfire SAM to verify the MAC on a buffer sent by a Desfire card
 *   (CLA=80 INS=5C) (functional specification 1.0, chapter 4.3.2)
 *
 * SYNOPSIS
 *   int DesfireSAMVerifyMAC(CARDAPPZ_CTX_ST     *ctx,
 *                   [input] const unsigned char Buffer[],
 *                           unsigned long       Length);
 *
 **/
int DesfireSAMVerifyMAC(CARDAPPZ_CTX_ST * ctx, const unsigned char Buffer[],
                        unsigned long Length)
{
  unsigned short status;
  int rc;

  unsigned long pos = 0;
  int len;

  int offset = 0;

  assert(ctx != NULL);
  if (Buffer == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  while (Length > pos)
  {
    /* We must send blocks of 0xFF bytes max */
    len = Length - offset;
    if (len > 0xFF)
    {
      len = 0xFF;
      rc =
        DesfireSamTransmit(ctx, 0x80, 0x5C, 0xAF, 0x00, &Buffer[pos], len,
                           -1, NULL, NULL, &status);
      if (rc)
        return rc;
      if (status != 0x90AF)
        return -status;         /* Wrong status */
    } else
    {
      rc =
        DesfireSamTransmit(ctx, 0x80, 0x5C, 0x00, 0x00, &Buffer[pos], len,
                           -1, NULL, NULL, &status);
      if (rc)
        return rc;
      if (status != 0x9000)
        return -status;         /* Wrong status */
    }
    pos += len;
  }

  /* OK */
  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMGenerateMAC
 *
 * NAME
 *   DesfireSAMGenerateMAC
 *
 * DESCRIPTION
 *   Use the Desfire SAM to compute the MAC on a buffer to be sent to a Desfire card
 *   (CLA=80 INS=7C) (functional specification 1.0, chapter 4.3.3)
 *
 * SYNOPSIS
 *   int DesfireSAMGenerateMAC(CARDAPPZ_CTX_ST     *ctx,
 *                    [input]  const unsigned char Buffer[],
 *                             int                 Length,
 *                    [output] unsigned char       Mac[4]);
 *
 **/
int DesfireSAMGenerateMAC(CARDAPPZ_CTX_ST * ctx,
                          const unsigned char Buffer[], unsigned long Length,
                          unsigned char Mac[4])
{
  unsigned char rsp[6];
  int rsplen;
  unsigned short status;
  int len;
  int rc;

  unsigned long pos = 0;

  assert(ctx != NULL);
  if (Buffer == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (Mac == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  while (Length > pos)
  {
    /* We must send blocks of 0xFF bytes max */
    len = Length - pos;
    if (len > 0xFF)
    {
      len = 0xFF;
      rsplen = -1;
      rc =
        DesfireSamTransmit(ctx, 0x80, 0x7C, 0xAF, 0x00, &Buffer[pos], len,
                           -1, NULL, NULL, &status);
    } else
    {
      /* Last frame */
      rsplen = 6;
      rc =
        DesfireSamTransmit(ctx, 0x80, 0x7C, 0x00, 0x00, &Buffer[pos], len, 0,
                           rsp, &rsplen, &status);
    }
    if (rc)
      return rc;
    pos += len;
  }

  if (rsplen != 6)
    return CARDAPPZ_CARD_PROTOCOL_ERROR;
  memcpy(Mac, rsp, 4);

  /* OK */
  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMEncipherData
 *
 * NAME
 *   DesfireSAMEncipherData
 *
 * DESCRIPTION
 *   Use the Desfire SAM to create a ciphered frame. CRC and padding is added,
 *   and Triple DES ciphering is performed.
 *   (CLA=80 INS=ED) (functional specification 1.0, chapter 4.3.5)
 *
 * SYNOPSIS
 *   int DesfireSAMEncipherData(CARDAPPZ_CTX_ST *ctx,
 *                     [in/out] unsigned char   Buffer[],
 *                              unsigned long   InLength,
 *                     [output] unsigned long   *OutLength);
 *
 **/
int DesfireSAMEncipherData(CARDAPPZ_CTX_ST * ctx, unsigned char Buffer[],
                           unsigned long InLength, unsigned long *OutLength)
{
  unsigned char rsp[0xF6 + 2];
  int rsplen;
  unsigned short status;
  int len;
  int rc;

  unsigned long pos = 0;

  assert(ctx != NULL);
  if (Buffer == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (OutLength == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (*OutLength < (8 + InLength))
    return CARDAPPZ_BUFFER_OVERFLOW;

  while (InLength > pos)
  {
    /* We must send blocks of 0xF6 bytes max */
    len = InLength - pos;
    rsplen = sizeof(rsp);
    if (len > 0xF6)
    {
      len = 0xF6;
      rc =
        DesfireSamTransmit(ctx, 0x80, 0xED, 0xAF, 0x00, &Buffer[pos], len, 0,
                           rsp, &rsplen, &status);
    } else
    {
      /* Last frame */
      rc =
        DesfireSamTransmit(ctx, 0x80, 0xED, 0x00, 0x00, &Buffer[pos], len, 0,
                           rsp, &rsplen, &status);
    }
    if (rc)
      return rc;
    if (rsplen < (2 + len))
      return CARDAPPZ_CARD_PROTOCOL_ERROR;
    if (rsplen > (10 + len))
      return CARDAPPZ_CARD_PROTOCOL_ERROR;

    memcpy(&Buffer[pos], rsp, rsplen - 2);
    pos += len;
  }

  /* OK */
  return CARDAPPZ_OK;
}

/**f* DesfireSAM/DesfireSAMDecipherData
 *
 * NAME
 *   DesfireSAMDecipherData
 *
 * DESCRIPTION
 *   Use the Desfire SAM to compute the MAC on a buffer to be sent to a Desfire card
 *   (CLA=80 INS=7C) (functional specification 1.0, chapter 4.3.3)
 *
 * SYNOPSIS
 *   int DesfireSAMDecipherData(CARDAPPZ_CTX_ST *ctx,
 *                     [in/out] unsigned char   Buffer[],
 *                              unsigned long   InLength,
 *                     [output] unsigned long   *OutLength);
 *
 **/
int DesfireSAMDecipherData(CARDAPPZ_CTX_ST * ctx, unsigned char Buffer[],
                           unsigned long InLength, unsigned long *OutLength)
{
  return CARDAPPZ_OK;
}

/*
 *****************************************************************************
 *
 * Desfire SAM human interface
 *
 *****************************************************************************
 */

int DesfireSamListKeys(CARDAPPZ_CTX_ST * ctx)
{
  unsigned char k;
  unsigned char e[12];
  int rc;

  printf("***** KEYS IN DESFIRE SAM *****\n");
  printf("#   : Va Vb Vc   Aid  Kid Ncek Vcek Nkuc  SET\n");

  /* We use only keys 2 to 9 */
  for (k = 0; k < 10; k++)
  {
    rc = DesfireSAMGetKeyEntry(ctx, k, e);
    if (rc)
    {
      printf("%03d : Err. %d\n", k, rc);
    } else
    {
      printf("%03d : ", k);
      printf("%02X %02X %02X ", e[0], e[1], e[2]);
      printf("%02X%02X%02X ", e[3], e[4], e[5]);
      printf(" %02X   %02X   %02X   %02X ", e[6], e[7], e[8], e[9]);
      printf("%02X%02X\n", e[10], e[11]);
    }
  }
  printf("\n");
  pause();

  return CARDAPPZ_NOT_YET_IMPLEMENTED;
}

int DesfireSamInstallKeys(CARDAPPZ_CTX_ST * ctx)
{
  int rc;

  rc = DesfireSAMAuthenticateHost(ctx, 0, DESFIRE_BLANK_KEY, NULL);
  if (rc)
    return rc;

  printf("Installing key 002...\n");
  /* Desfire mode */
  rc =
    DesfireSAMChangeKeyEntry(ctx, 2, 0x82, CORPID_MASTER_KEY0, NULL, NULL, 0,
                             0, 0, 0, 0xFF, 0x0000, NULL);
  if (rc)
    return rc;

  printf("Installing key 003...\n");
  /* Desfire mode */
  rc =
    DesfireSAMChangeKeyEntry(ctx, 3, 0x82, CORPID_MASTER_KEY1, NULL, NULL, 0,
                             0, 0, 0, 0xFF, 0x0000, NULL);
  if (rc)
    return rc;

  printf("Installing key 004...\n");
  /* ISO mode, allow crypto with secret key (SAM UID must be provided) */
  rc =
    DesfireSAMChangeKeyEntry(ctx, 4, 0x82, CORPID_SIGNATURE_KEY, NULL, NULL,
                             0, 0, 0, 0, 0xFF, 0x000B, NULL);
  if (rc)
    return rc;

  printf("Installing key 005...\n");
  /* Desfire mode */
  rc =
    DesfireSAMChangeKeyEntry(ctx, 5, 0x82, PURSE_MASTER_KEY0, NULL, NULL, 0,
                             0, 0, 0, 0xFF, 0x0000, NULL);
  if (rc)
    return rc;

  printf("Installing key 006...\n");
  /* Desfire mode */
  rc =
    DesfireSAMChangeKeyEntry(ctx, 6, 0x82, PURSE_MASTER_KEY1, NULL, NULL, 0,
                             0, 0, 0, 0xFF, 0x0000, NULL);
  if (rc)
    return rc;

  printf("Installing key 007...\n");
  /* Desfire mode */
  rc =
    DesfireSAMChangeKeyEntry(ctx, 7, 0x82, TRANSPORT_MASTER_KEY0, NULL, NULL,
                             0, 0, 0, 0, 0xFF, 0x0000, NULL);
  if (rc)
    return rc;

  printf("Installing key 008...\n");
  /* Desfire mode */
  rc =
    DesfireSAMChangeKeyEntry(ctx, 8, 0x82, TRANSPORT_MASTER_KEY1, NULL, NULL,
                             0, 0, 0, 0, 0xFF, 0x0000, NULL);
  if (rc)
    return rc;

  printf("Installing key 009...\n");
  /* ISO mode, allow crypto with secret key (SAM UID must be provided) */
  rc =
    DesfireSAMChangeKeyEntry(ctx, 9, 0x82, TRANSPORT_SIGNATURE_KEY, NULL,
                             NULL, 0, 0, 0, 0, 0xFF, 0x000B, NULL);
  if (rc)
    return rc;

  return CARDAPPZ_OK;
}

int DesfireSamBlankKeys(CARDAPPZ_CTX_ST * ctx)
{
  unsigned char k;
  unsigned char v[16] = { 0 };
  int rc;

  rc = DesfireSAMAuthenticateHost(ctx, 0, DESFIRE_BLANK_KEY, NULL);
  if (rc)
    return rc;

  /* We drop keys 1 to 9... */
  for (k = 1; k < 10; k++)
  {
    printf("Blanking key %03d... ", k);
    rc =
      DesfireSAMChangeKeyEntry(ctx, k, 0xFE, v, v, v, 0, 0, 0, 0, 0xFF,
                               0x0000, NULL);
    if (rc)
      printf("Err. %d\n", rc);
    else
      printf("OK\n");
  }

  return CARDAPPZ_OK;
}

int DesfireSamTest(CARDAPPZ_CTX_ST * ctx)
{
  unsigned char bufS[16], bufH[8];
  unsigned long l;
  int i, j;
  int rc;
  TDES_CTX_ST tdes_ctx;
  unsigned char SamUid[7];

  rc = DesfireSAMGetVersion(ctx, NULL, SamUid);
  if (rc)
    return rc;

  rc = DesfireSAMAuthenticateHost(ctx, 0, DESFIRE_BLANK_KEY, NULL);
  if (rc)
    return rc;

  rc = DesfireSAMAuthenticateHost(ctx, 0, DESFIRE_BLANK_KEY, SamUid);
  if (rc)
    return rc;

  TDES_Init(&tdes_ctx, &TRANSPORT_SIGNATURE_KEY[0],
            &TRANSPORT_SIGNATURE_KEY[8], &TRANSPORT_SIGNATURE_KEY[0]);

  rc =
    DesfireSAMSelectKey(ctx, TRANSPORT_SIGNATURE_KEY_IDX,
                        TRANSPORT_SIGNATURE_KEY_VER);
  if (rc)
    return rc;

  for (i = 0; i < 255; i++)
  {
    memset(bufS, i, 8);
    memset(bufH, i, 8);

    TDES_Encrypt(&tdes_ctx, bufH);

    l = 16;
    rc = DesfireSAMEncipherData(ctx, bufS, 8, &l);
    if (rc)
    {
      printf("SAM encipher failed %d\n", rc);
      continue;
    }

    if (memcmp(bufS, bufH, 8))
    {
      printf("Encipher error...\n");
      for (j = 0; j < 8; j++)
        printf("%02X", bufH[i]);
      printf(" != ");
      for (j = 0; j < 16; j++)
        printf("%02X", bufS[i]);
      printf("\n");
    }

    l = 16;
    rc = DesfireSAMDecipherData(ctx, bufS, 16, &l);
    if (rc)
    {
      printf("SAM decipher failed %d\n", rc);
      continue;
    }
    if (l != 8)
      printf("Oups decipher wrong length : %d\n", l);

    TDES_Decrypt(&tdes_ctx, bufH);

    if (memcmp(bufS, bufH, 8))
    {
      printf("Decipher error...\n");
      for (j = 0; j < 8; j++)
        printf("%02X", bufH[i]);
      printf(" != ");
      for (j = 0; j < 16; j++)
        printf("%02X", bufS[i]);
      printf("\n");
    }
  }


  return CARDAPPZ_NOT_YET_IMPLEMENTED;
}

int DesfireSamMenu(CARDAPPZ_CTX_ST * ctx)
{
  int rc;
  int c;

  for (;;)
  {
    printf("\n");
    printf("***** DESFIRE SAM MENU *****\n");
    printf("L -- (L)ist installed keys\n");
    printf("I -- (I)nstall our keys\n");
    printf("B -- (B)lank the card (erase all keys)\n");
    printf("T -- (T)est the implementation\n");
    printf("\n");
    printf("X -- E(X)it - Back to main menu\n");

    c = toupper(getch());
    switch (c)
    {
      case 'L':
        rc = DesfireSamListKeys(ctx);
        break;
      case 'I':
        rc = DesfireSamInstallKeys(ctx);
        break;
      case 'B':
        rc = DesfireSamBlankKeys(ctx);
        break;
      case 'T':
        rc = DesfireSamTest(ctx);
        break;
      case 'X':
        return CARDAPPZ_OK;
    }
  }

  return CARDAPPZ_OK;
}
