/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2003-2009 PRO ACTIVE SAS - www.proactive-group.com
   
  Ref_Mifare.c
  ------------

  This is the reference application that shows how to read/write data
  in a Philips MIFARE classic tag.
  
  JDA 21/11/2003 : initial release
  JDA 03/11/2006 : calling SPROX_A_SelectIdle and SPROX_A_Halt instead
                   of Mf500xxxx functions
  JDA 10/01/2007 : fixed a few warnings when compiling under GCC
  JDA 23/02/2007 : added the '-b' flag for "benchmarking"

*/

#include "springprox.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

int main(int argc, char **argv)
{
  SWORD rc;
  int i;
  char s_buffer[64];
  BOOL bench = FALSE;
  clock_t t0, t1;
  BYTE data[240];
  BYTE atq[2];
  BYTE sak[1];
  BYTE uid[12];
  BYTE uid_len = 12;

  BYTE sect_count, sect;
  BYTE bloc_count, bloc;
  BYTE offset;

  char *device = NULL;
  
  i = 1;
  if ((argc > i) && !strcmp(argv[i], "-b"))
  {
    bench = TRUE;
    i++;
  }
  if (argc > i)
    device = argv[i];

  printf("SpringCard SDK for K531/SpringProx-CF/CSB Legacy\n");
  printf("\n");
  printf("NXP MIFARE 'CLASSIC' (1k and 4k) reference demo\n");
  printf("-----------------------------------------------\n");
  printf("www.springcard.com\n\n");

  rc = SPROX_GetLibrary(s_buffer, sizeof(s_buffer));
  if (rc != MI_OK)
  {
    printf("Failed to get API version\n");
    goto done;
  }
  printf("API version : %s\n", s_buffer);
  printf("\n");

  /* Open reader */
  /* ----------- */

  rc = SPROX_ReaderOpen(device);
  if (rc != MI_OK)
  {
    printf("Reader not found\n");
    goto done;
  }

  rc = SPROX_ReaderGetDevice(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader found on %s\n", s_buffer);

  rc = SPROX_ReaderGetFirmware(s_buffer, sizeof(s_buffer));
  if (rc == MI_OK)
    printf("Reader firwmare is %s\n", s_buffer);

  /* Configure reader in ISO 14443-A mode */
  /* ------------------------------------ */
  rc = SPROX_SetConfig(CFG_MODE_ISO_14443_A);
  if (rc != MI_OK)
  {
    printf("Failed to configure the reader\n");
    goto close;
  }

  /* RF field ON */
  SPROX_ControlRF(TRUE);

  /* Select any tag in the ISO 14443-A RF field */
  /* ------------------------------------------ */
  rc = SPROX_A_SelectIdle(atq, uid, &uid_len, sak);
  if (rc != MI_OK)
  {
    printf("No available tag in RF field\n");
    goto close;
  }

  printf("Tag found: UID=");
  for (i = 0; i < uid_len; i++)
    printf("%02X", uid[i]);
  printf(" ATQ=%02X%02X SAK=%02X\n", atq[1], atq[0], sak[0]);

  if ((atq[1] != 0x00)
      || ((atq[0] != 0x02) && (atq[0] != 0x04) && (atq[0] != 0x18)))
  {
    printf("This is not a Mifare classic tag\n");
    goto tag_halt;
  }

	sect_count = 0;
  if ((sak[0] & 0x1F) == 0x08)
  {
    /* Mifare classic 1k : 16 sectors, 3+1 blocks in each sector */
    printf("Tag appears to be a Mifare classic 1k\n");
    sect_count = 16;
  } else if ((sak[0] & 0x1F) == 0x18)
  {
    /* Mifare classic 4k : 40 sectors, 3+1 blocks in 32-first sectors, 15+1 blocks in the 8-last sectors */
    printf("Tag appears to be a Mifare classic 4k\n");
    sect_count = 40;
  }
  
  if (bench)
  {
    printf("Reading %d sectors...\n", sect_count);
    t0 = clock();
  }

  bloc = 0;
  for (sect = 0; sect < sect_count; sect++)
  {
    if (!bench)
      printf("Reading sector %02d : ", sect);

    /* The SPROX_MifStReadSector functions assumes that the data buffer is big enough... */
    rc = SPROX_MifStReadSector(NULL, (BYTE) sect, data, NULL);
    if (rc != MI_OK)
    {
      if (bench)
        printf("Read sector %02d ", sect);
      printf("[Failed]\n");
      printf("  %s (%d)\n", SPROX_GetErrorMessage(rc), rc);
    } else
    if (!bench)
    {
      printf("[OK]\n");
      /* Display sector's data */
      if (sect < 32)
        bloc_count = 3;
      else
        bloc_count = 15;
      for (bloc = 0; bloc < bloc_count; bloc++)
      {
        printf("%02d : ", bloc);
        /* Each blocks is 16-bytes wide */
        for (offset = 0; offset < 16; offset++)
        {
          printf("%02X ", data[16 * bloc + offset]);
        }
        for (offset = 0; offset < 16; offset++)
        {
          if (data[16 * bloc + offset] >= ' ')
          {
            printf("%c", data[16 * bloc + offset]);
          } else
          {
            printf(".");
          }
        }
        printf("\n");
      }
    }
  }
  
  if (bench)
  {
    t1 = clock(); 
    printf("Time elapsed: %ldms\n", (t1 - t0) / (CLOCKS_PER_SEC/1000));
  }

  goto tag_halt;

tag_halt:
  /* Halt the tag */
  rc = SPROX_A_Halt();
  if (rc != MI_OK)
  {
    printf("Failed to halt the tag\n");
    goto close;
  }

close:
  /* Close the reader */
  SPROX_ControlRF(FALSE);
  SPROX_ReaderClose();

done:
  /* Display last error */
  if (rc == MI_OK)
  {
    printf("Done\n");
  } else
  {
    printf("%s (%d)\n", SPROX_GetErrorMessage(rc), rc);
  }
  return 0;
}
